#ifndef INSTANCE_TREE_H_INCLUDED_
#define INSTANCE_TREE_H_INCLUDED_

#include "rose.h"

#include <vector>
#include <list>
#include <utility>

namespace risc {

class Class;
class Object;

class InstanceTree {

public:
  InstanceTree();

  InstanceTree(const InstanceTree &other);

  struct Element {
    Class* first;
    Object* second;
    int id;

    Element(Class* first_, Object* second_, int id_ = 0):
      first(first_), second(second_), id(id_)
    { }

    Element():
      first(NULL), second(NULL), id(0)
    { }
  };

  typedef std::list<Element> Path;

  /**
   * \fn       get_real_mapped_variable
   * \brief    This function returns mapped variables for a reference.
   *           If the mapped variable is a global variable, the first scope
   *           pointer will be NULL
   */
  InstanceTree get_real_mapped_variable();

  /**
   * \fn       find_all_instances_of
   * \brief    This function returns all variables
   */
  std::list<InstanceTree> find_all_instances_of(SgVariableDefinition *var_def);

  /**
   * \fn       get_parent
   * \brief    This function returns the parent variable of variable
   */
  InstanceTree get_parent();

  /**
   * \fn       print_instance_tree
   * \brief    Prints the instance tree
   */
  void print_instance_tree();

  /**
   * \fn       get_scope
   * \brief    returns the scope of the object
   */
  Class* get_scope();

  /**
   * \fn       get_object
   * \brief    returns the object
   */
  Object* get_object();

  /**
   * \fn       get_object_symbol
   * \brief    returns the symbol of the object
   */
  SgVariableSymbol* get_object_symbol();

  /**
   * \fn       get_all_ports
   * \brief    returns all ports
   */
  void get_all_ports(std::list<InstanceTree>& occurence);

  /**
   * \fn       get_all_events
   * \brief    returns all events
   */
  void get_all_events(std::list<InstanceTree>& occurence);

  /**
   * \fn       pop_back
   * \brief    Deletes the last element in the instance tree
   */
  Element pop_back();

  /**
   * \fn       push_back
   * \brief    Appends element ohs at the end of the list
   */
  void push_back(Element ohs);

  /**
   * \brief These are the kind for of variables for which a traversal can search
   */
  enum VariableType {
    ModuleInstance,
    PrimitiveChannelInstance,
    HierarchicalChannelInstance,
    EventInstance,
    VariableInstance,
    PortInstance
  };

  /**
   * \fn                get_instance_tree
   * \brief             find the instance path to a variable
   *                    with a given instance id of the hosting module
   * \param vardef      variable for which is searched
   * \param instance_id instance id of the hosting module/channel
   * \param tree        search starts from here
   * \param type        type of the variable
   * \info              Caution: It is not possible to search for variables
   *                             in channels!
   */
  void
  get_instance_tree(
      SgVariableDefinition *var_def,
      int instance_id,
      risc::InstanceTree &tree,
      VariableType type);

  /**
   * \fn                recursive_tree_walk
   * \brief             For internal usage only!
   */
  void
  recursive_tree_walk(
      SgVariableDefinition *var_def,
      int &current_id,
      int instance_id,
      risc::InstanceTree &tree,
      bool &found,
      VariableType type);

  bool operator==(const InstanceTree &ohs);

  Path path_;

private:
  /**
   * \fn       recursive_tree_walk
   * \brief    Iterates recursively over the tree and collects all matching
   *           objects
   */
  void recursive_tree_walk(std::list<InstanceTree>& occurence,
           SgVariableDefinition *var_def);

  /**
   * \fn       get_mapped_reference
   * \brief    This function determines the mapped variables symbol in the
   *           parent class.
   *           Only for internal use
   */
  SgVariableSymbol* get_mapped_reference(SgVariableSymbol* reference,
    SgClassDefinition *base_class_def,
    SgClassDefinition *parent_class_def,
    SgVariableDefinition *parent_member);
};

} // end of namespace risc

#endif /* INSTANCE_TREE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
