#include "port.h"

#include "inport.h"
#include "inoutport.h"
#include "outport.h"

#include "instance.h"
#include "module_instance.h"
#include "primitive_channel_instance.h"
#include "hierarchical_channel_instance.h"

risc::Port::Port(SgVariableDefinition *ast_node, Class *parent, Type type):
  Object(ast_node, type),
  parent_(parent),
  bounded (false)
{ }

std::string
risc::Port::get_ast_type_name()
{
  return get_ast_type()->unparseToString().substr(17);
}

void risc::Port::get_mapped_channels(std::set<risc::Object*> &channels)
{
  // Strategy
  // One port can be mapped to many channels
  // because we are on the level of a module definition.
  //
  // So, as a starting point we have a port.
  // First, we identify the surrounding module/channel where it is defined.
  // The module/channel definition stores a list of all its instances.
  // For each instance we look up the mapping of the port.
  // From there on go recursively down until there is a mapping to a channel.

  // We don't need to cast this one to a hierarchical channel as well
  // because the class Module derives the instances list
  // to the definition of hierarchical channels
  Module *parent = dynamic_cast<Module*>(parent_);

  for(std::list<Instance*>::iterator
      instance_iter  = parent->instances_.begin();
      instance_iter != parent->instances_.end();
      instance_iter++) {

    ModuleInstance *module_instance = NULL;
    if((*instance_iter)->type_ == ModuleInstanceType) {
      module_instance = dynamic_cast<ModuleInstance*>(*instance_iter);
    }

    HierarchicalChannelInstance *hierarchical_channel_instance = NULL;
    if((*instance_iter)->type_ == HierarchicalChannelInstanceType) {
      hierarchical_channel_instance
        = dynamic_cast<HierarchicalChannelInstance*>(*instance_iter);
    }

    std::map<Port*, Object*>::iterator mapped;

    if(module_instance) {
      mapped = module_instance->mapping_.find(this);
    }
    if(hierarchical_channel_instance) {
      mapped = hierarchical_channel_instance->mapping_.find(this);
    }

    InPort *inport = dynamic_cast<InPort*>((*mapped).second);
    InOutPort *inoutport = dynamic_cast<InOutPort*>((*mapped).second);
    OutPort *outport = dynamic_cast<OutPort*>((*mapped).second);

    if(inport) {

      inport->get_mapped_channels(channels);
    } else if(inoutport) {

      inoutport->get_mapped_channels(channels);
    } else if(outport) {

      outport->get_mapped_channels(channels);
    } else {

      channels.insert((*mapped).second);
    }
  }
}

void risc::Port::get_mapped_channel_types(std::set<risc::Class*> &channels)
{
  // Strategy
  // One port can be mapped to many channels
  // because we are on the level of a module definition.
  //
  // So, as a starting point we have a port.
  // First, we identify the surrounding module/channel where it is defined.
  // The module/channel definition stores a list of all its instances.
  // For each instance we look up the mapping of the port.
  // From there on go recursively down until there is a mapping to a channel.

  // We don't need to cast this one to a hierarchical channel as well
  // because the class Module derives the instances list
  // to the definition of hierarchical channels
  Module *parent = dynamic_cast<Module*>(parent_);

  for(std::list<Instance*>::iterator
      instance_iter  = parent->instances_.begin();
      instance_iter != parent->instances_.end();
      instance_iter++) {

    ModuleInstance *module_instance = NULL;
    if((*instance_iter)->type_ == ModuleInstanceType) {
      module_instance = dynamic_cast<ModuleInstance*>(*instance_iter);
    }

    HierarchicalChannelInstance *hierarchical_channel_instance = NULL;
    if((*instance_iter)->type_ == HierarchicalChannelInstanceType) {
      hierarchical_channel_instance
        = dynamic_cast<HierarchicalChannelInstance*>(*instance_iter);
    }

    std::map<Port*, Object*>::iterator mapped;

    if(module_instance) {
      mapped = module_instance->mapping_.find(this);
    }
    if(hierarchical_channel_instance) {
      mapped = hierarchical_channel_instance->mapping_.find(this);
    }

    InPort *inport = dynamic_cast<InPort*>((*mapped).second);
    InOutPort *inoutport = dynamic_cast<InOutPort*>((*mapped).second);
    OutPort *outport = dynamic_cast<OutPort*>((*mapped).second);

    if(inport) {

      inport->get_mapped_channel_types(channels);
    } else if(inoutport) {

      inoutport->get_mapped_channel_types(channels);
    } else if(outport) {

      outport->get_mapped_channel_types(channels);
    } else {

      PrimitiveChannelInstance *prim_inst
        = dynamic_cast<PrimitiveChannelInstance*>((*mapped).second);

      HierarchicalChannelInstance *hier_inst
        = dynamic_cast<HierarchicalChannelInstance*>((*mapped).second);

      if(prim_inst) {

        assert(prim_inst->get_definition() && "No channel definition available!");
        channels.insert(prim_inst->get_definition());

      } else if(hier_inst) {

        assert(hier_inst->get_definition() && "No channel definition available!");
        channels.insert(hier_inst->get_definition());

      } else {

        assert(0 && "No channel definition available!");
      }
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
