#include "conflict_table.h"

#include<vector>
#include<utility>
#include<iomanip>

#include "graph_helper.h"

#include "../tools/misc.h"
#include "../tools/function_helper.h"
#include "../tools/variable_helper.h"
#include "../segment_graph/reference_attributes.h"
#include "../internal_representation/cthread.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/hierarchical_channel_instance.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/method.h"
#include "../internal_representation/module.h"
#include "../internal_representation/module_instance.h"
#include "../internal_representation/path_instance_mapper.h"
#include "../internal_representation/port_mapping.h"
#include "../internal_representation/primitive_channel.h"
#include "../internal_representation/primitive_channel_instance.h"
#include "../internal_representation/static_path_instance_mapper.h"
#include "../internal_representation/thread.h"

risc::sg::MappedVariableList
risc::sg::ConflictTable::get_mapped_variable(SgVariableSymbol *variable,
    int instance_id, bool dynamic_analysis)
{
  if(dynamic_analysis) {

    return get_mapped_variable_dynamically(variable, instance_id);
  } else {

    return get_mapped_variable_statically(variable, instance_id);
  }
}

risc::sg::MappedVariableList
risc::sg::ConflictTable::get_mapped_variable_dynamically(
    SgVariableSymbol *variable,
    int instance_id)
{
  if(isSgReferenceType(variable->get_type())) {
    SgClassDefinition *enclosing_class_def_of_variable
      = SageInterface::getEnclosingClassDefinition(variable);

    std::string host_module_type_name
      = risc::tools::get_mangled_name(enclosing_class_def_of_variable);
    host_module_type_name.erase(0 , 6 );


    // Variable name without __symbol__
    std::string reference_name = SageInterface::get_name(variable);
    reference_name.erase(reference_name.length() - 8 , 8 );

    // Here we determine to which variable the reference is mapped
    int counter = 0;
    int *mapped_variable_address = get_dynamic_object_of_variable(
      path_instance_mapper_->get_dynamic_tree_root(),
      host_module_type_name,
      instance_id,
      reference_name,
      counter)->address_;

    // Searching for the host module and the variable name of the variable
    // address
    std::pair<DynamicModule*, std::string> host_module_and_variable
      = get_host_module_and_name_of_variable(
          path_instance_mapper_->get_dynamic_tree_root(),
          mapped_variable_address);

    // If the host module pointer is NULL it must be a global variable
    if(host_module_and_variable.first == NULL) {

      // So, we are searching for the name of the variable in the global scope
      std::string mapped_variable
        = path_instance_mapper_->get_global_variable_name(
                                     mapped_variable_address);

      // Finally, we are translating the variable name into a variable symbol
      SgVariableSymbol *mapped_variable_symbol
        = risc::tools::find_global_variable(
                           mapped_variable,
                           risc::get_project());

      // the value -1 indicates that this variable is a global variable
      MappedVariable temp(mapped_variable_symbol, -1);
      MappedVariableList result;
      result.push_back(temp);
      return result;
    }

    // If we are reaching this point, it is a member variable
    // So, next we determine the instance id of the hosting module of the
    // variable
    int mapped_instance_id = 0;
    bool result = get_instance_id_of_hosting_module(
                      path_instance_mapper_->get_dynamic_tree_root(),
                      mapped_variable_address,
                      host_module_and_variable.first->name_,
                      mapped_instance_id);

    // At this point, we have to find the associated module definition
    // in the internal representation. We can do this by comparing the
    // modules types (mangled names)
    risc::Module *current_module
      = risc::find_module_definition(host_module_and_variable.first->name_);

    SgClassDefinition *class_def = NULL;

    if(current_module) {
      class_def = current_module->get_ast_node();
    } else {
      risc::HierarchicalChannel *current_hierarchical_channel
        = risc::find_hierarchical_channel_definition(
                    host_module_and_variable.first->name_);

      class_def = current_hierarchical_channel->get_ast_node();
    }

    SgVariableSymbol *mapped_symbol = NULL;

    // After having the module, we have to search for the mapped variable in the
    // class. So, we are iterating over all the declarations in the class and
    // compare all the variables names with name of the mapped variable
    for(SgDeclarationStatementPtrList::iterator
        iter  = class_def->get_members().begin();
        iter != class_def->get_members().end();
        iter++) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      if(var_decl) {

        std::string temp_variable_name(SageInterface::get_name(var_decl));

        // deleting __variable_declaration__ out of the string
        temp_variable_name.erase(0, 22);

        if(temp_variable_name == host_module_and_variable.second) {
          mapped_symbol = isSgVariableSymbol(var_decl->get_decl_item(0)
              ->search_for_symbol_from_symbol_table());
        }
      }
    }

    risc::sg::MappedVariableList return_value;
    return_value.push_back(MappedVariable(mapped_symbol, mapped_instance_id));
    return return_value;

  } else {

    risc::sg::MappedVariableList result;
    result.push_back(MappedVariable(variable, instance_id));
    return result;
  }
}

risc::sg::MappedVariableList
risc::sg::ConflictTable::get_mapped_variable_statically(
    SgVariableSymbol *variable,
    int instance_id)
{
  risc::sg::MappedVariableList result;

  // Here, we check if the variable belongs to a channel.
  // We don't associate channels with instance id's.
  // So, every channel variable gets the instance id 0.
  if(tools::is_channel_variable(variable)) {

    MappedVariable temp(variable, 0);
    result.push_back(temp);
    return result;
  }

  if(isSgReferenceType(variable->get_type())) {

    if(tools::is_member_variable(variable)) {
      SgVariableDefinition *var_def
        = isSgVariableDefinition(variable->get_declaration()
            ->get_definition());

      std::list<InstanceTree> potential_instances
        = risc::get_design()->get_root().find_all_instances_of(var_def);

      InstanceTree waiting_event_tree
        = path_instance_mapper_->get_instance_tree(
            instance_id, potential_instances);

      // in this case there was no mapping.
      if(waiting_event_tree.path_.empty()) {
        return result;
      }

      std::list<risc::InstanceTree> temp
        = waiting_event_tree.find_all_instances_of(var_def);

      risc::InstanceTree mapped_variable
        = temp.front().get_real_mapped_variable();

      // in this case it is a global variable
      if(mapped_variable.path_.size() == 1) {
        if(mapped_variable.path_.front().first == NULL) {

          result.push_back(
                     MappedVariable(mapped_variable.get_object_symbol(), -1));
          return result;
        }
      }

      risc::InstanceTree hosting_module = mapped_variable;
      hosting_module.pop_back();

      result.push_back(
                 MappedVariable(
                     mapped_variable.get_object_symbol(),
                     path_instance_mapper_->get_id(hosting_module)));
      return result;

    } else if(tools::is_function_parameter_variable(variable)) {

      FunctionParameterReferenceAttribute *parameter_variable =
        dynamic_cast<FunctionParameterReferenceAttribute*>(
            variable->getAttribute("parameter_reference"));

      bool is_channel_variable = tools::is_channel_variable(variable);

      for(std::set<SgVariableSymbol*>::iterator
          iter  = parameter_variable->mapped_variables_.begin();
          iter != parameter_variable->mapped_variables_.end();
          iter++) {

        result.push_back(MappedVariable(*iter, instance_id));
      }

    } else if(tools::is_local_variable(variable)) {

      risc::sg::LocalReferenceAttribute *local_variable
        = dynamic_cast<risc::sg::LocalReferenceAttribute*>(
            variable->getAttribute("local_reference"));

      SgVariableSymbol *variable_symbol = local_variable->mapped_variable_;

      if(tools::is_local_variable(variable_symbol)) {
        result.push_back(MappedVariable(
                             local_variable->mapped_variable_, instance_id));

      } else if(tools::is_function_parameter_variable(variable_symbol)) {

        FunctionParameterReferenceAttribute *parameter_variable =
          dynamic_cast<FunctionParameterReferenceAttribute*>(
              variable_symbol->getAttribute("parameter_reference"));

        for(std::set<SgVariableSymbol*>::iterator
            iter  = parameter_variable->mapped_variables_.begin();
            iter != parameter_variable->mapped_variables_.end();
            iter++) {

          result.push_back(MappedVariable(*iter, instance_id));
        }

      } else if(tools::is_member_variable(variable_symbol)) {
        MappedVariableList temp
          = get_mapped_variable_statically(variable, instance_id);
        result.insert(result.end(), temp.begin(), temp.end());

      } else {
        assert(0 && "No support for solving this reference");
      }

      return result;

    } else {

      assert(0 && "No support for solving this reference");
    }
  } else {

    result.push_back(MappedVariable(variable, instance_id));
    return result;
  }
}


risc::sg::ConflictTable::ConflictTable(
    SegmentGraph &segment_graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  segment_graph_(segment_graph),
  path_instance_mapper_(path_instance_mapper),
  debugging_(debugging)
{
  number_of_segments_ = boost::num_vertices(segment_graph_.graph_);
  max_instances_ = path_instance_mapper_->max_instances_;

  instance_id_to_index_lookup_ = new int*[max_instances_ + 1];

  // We are adding one additional row to safe the segment id
  for(int i=0; i < max_instances_ + 1; i++) {

    instance_id_to_index_lookup_[i] = new int[number_of_segments_];
  }

  // Here, the first row of the instance_id_to_index_lookup table is filled
  // with the segment ids. We have to do that because the segment ids do not
  // incease by one.
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    instance_id_to_index_lookup_[0][i] = segment_graph_.graph_[*(vp.first)].id_;
  }

  // We are filling all array fields with -1 to indicate that a specific field
  // is not used. We need that information for the regression tests.
  for(int i = 1; i < max_instances_ + 1; i++) {
    for(int j=0; j < number_of_segments_; j++) {
      instance_id_to_index_lookup_[i][j] = -1;
    }
  }

  // Here, we are filling the instance_id_to_index_lookup_ table. Each field
  // in this table needs an unique value. The value is the associated row/column
  // in the conflict table.
  int index = 0;

  // We are iterating over all the instances in the design
  for(std::vector<PathInstanceMapper::PathWithID>::iterator
      iter  = path_instance_mapper_->path_to_id_mapping_.begin();
      iter != path_instance_mapper_->path_to_id_mapping_.end();
      iter++) {

    // Adding the value of 1 because the first row is reserved for the
    // segment id
    int row = (*iter).second + 1;

    Object *object = (*iter).first.get_object();

    ModuleInstance *module_instance = dynamic_cast<ModuleInstance*>(object);

    Module *module = NULL;

    if(module_instance) {
      module = module_instance->get_definition();
    }

    if(!module) {

      module = dynamic_cast<Module*>((*iter).first.path_.back().first);
    }

    if(module) {

      for(ThreadVectorConstIter
          thread_iter  = module->threads_.begin();
          thread_iter != module->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorConstIter
          cthread_iter  = module->cthreads_.begin();
          cthread_iter != module->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorConstIter
          method_iter  = module->methods_.begin();
          method_iter != module->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }

    // This is the same like for the modules. So, please see the documents above
    HierarchicalChannelInstance *hierarchical_channel_instance
      = dynamic_cast<HierarchicalChannelInstance*>(object);

    if(hierarchical_channel_instance) {

      HierarchicalChannel *hierarchical_channel
        = hierarchical_channel_instance->get_definition();

      for(ThreadVectorConstIter
          thread_iter  = hierarchical_channel->threads_.begin();
          thread_iter != hierarchical_channel->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorConstIter
          cthread_iter  = hierarchical_channel->cthreads_.begin();
          cthread_iter != hierarchical_channel->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorConstIter
          method_iter  = hierarchical_channel->methods_.begin();
          method_iter != hierarchical_channel->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }
  }

  size_of_conflict_table_ = index;


  // Allocating the memory for the VertexDescriptor array
  vertex_lookup_ = new VertexDescriptor[number_of_segments_];

  // Here, we are filling the vertex lookup. The array field at position is the
  // segment with the index instance_id_to_index_lookup_[0][i]
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    int i = 0;
    for(; i < number_of_segments_; i++) {
      if(instance_id_to_index_lookup_[0][i]
          == segment_graph_.graph_[*(vp.first)].id_) {

        vertex_lookup_[i] = *(vp.first);
        break;
      }
    }
  }

  // creating the conflict table
  conflict_table_ = new bool*[size_of_conflict_table_];
  for(int i=0; i<size_of_conflict_table_; i++) {
    conflict_table_[i] = new bool[size_of_conflict_table_];
  }

  if(debugging_) {

    conflict_info_ = new ConflictInformation*[size_of_conflict_table_];
    for(int i=0; i<size_of_conflict_table_; i++) {
      conflict_info_[i] = new ConflictInformation[size_of_conflict_table_];
    }
  } else {

    conflict_info_ = NULL;
  }

  max_instances_with_thread_ = 0;
  for(int i = 0; i < boost::num_vertices(segment_graph_.graph_); i++) {
    for(int j = 0; j < max_instances_; j++) {

      if(instance_id_to_index_lookup_[j][i] == -1) {
        break;
      } else {
        if(j > max_instances_with_thread_) {
          max_instances_with_thread_ = j;
        }
      }
    }
  }

  if(debugging_ && !path_instance_mapper_->perform_dynamic_analysis()) {
    conflicts_per_instance_
      = new int [max_instances_ * max_instances_];
    for(int i = 0;
            i < max_instances_ * max_instances_;
            i++) {
      conflicts_per_instance_[i] = 0;
    }
  }
}

void
risc::sg::ConflictTable::determine_conflict_table()
{
  for(int i=0; i<size_of_conflict_table_; i++) {
    for(int j=0; j<size_of_conflict_table_; j++) {

      conflict_table_[i][j] = false;
    }
  }
}

risc::sg::ConflictTable::~ConflictTable()
{
  for(int i=0; i < max_instances_ + 1; i++) {

    delete[] instance_id_to_index_lookup_[i];
  }
  delete[] instance_id_to_index_lookup_;


  for(int i=0; i < size_of_conflict_table_; i++) {

    delete[] conflict_table_[i];
  }

  delete[] conflict_table_;

  delete[] vertex_lookup_;

  if(debugging_) {
    for(int i=0; i < size_of_conflict_table_; i++) {

      delete[] conflict_info_[i];
    }
    delete[] conflict_info_;
  }

  if(debugging_ && !path_instance_mapper_->perform_dynamic_analysis()) {
    delete conflicts_per_instance_;
  }
}

void
risc::sg::ConflictTable::print_instance_id_to_index_lookup_table()
{

  for(int i=0; i < max_instances_ + 1; i++) {

    if(i == 0) {

      std::cout << "    ";
    } else {
      std::cout << std::setw(4) << i-1;
    }

    for(int j=0; j < boost::num_vertices(segment_graph_.graph_); j++) {

      std::cout << std::setw(4) << instance_id_to_index_lookup_[i][j] << " ";
    }

    std::cout << std::endl;
  }
}

int
risc::sg::ConflictTable::segment_and_instance_id_to_index(
    int segment_id, int instance_id)
{
  int column_index = -1;
  for(int i = 0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id) {

      column_index = i;
    }
  }

  if(column_index == -1) {
    assert("segment_and_instance_id_to_index: This segment id does not exist"
        && 0);
  }

  return instance_id_to_index_lookup_ [instance_id + 1][column_index];
}

std::pair<int, int>
risc::sg::ConflictTable::index_to_segment_and_instance_id(int index)
{
  std::pair<int, int> result;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      if(instance_id_to_index_lookup_[i][j] == index) {

        result.first = instance_id_to_index_lookup_[0][j];
        result.second = i - 1;
        return result;
      }
    }
  }

  std::cout << "index: " << index << std::endl;
  assert("index_to_segment_and_instance_id: Index does not exist" && 0);
}

bool
risc::sg::ConflictTable::has_conflict(int segment_id_1, int instance_id_1,
    int segment_id_2, int instance_id_2)
{
  return conflict_table_
    [segment_and_instance_id_to_index(segment_id_1, instance_id_1)]
    [segment_and_instance_id_to_index(segment_id_2, instance_id_2)];
}

void
risc::sg::ConflictTable::print_conflict_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;

  myfile << ".tg .tg-red0{background-color:#fe0000}" << std::endl;
  myfile << ".tg .tg-red1{background-color:#ee0000}" << std::endl;
  myfile << ".tg .tg-red2{background-color:#de0000}" << std::endl;
  myfile << ".tg .tg-green0{background-color:#009900}" << std::endl;
  myfile << ".tg .tg-green1{background-color:#008900}" << std::endl;
  myfile << ".tg .tg-green2{background-color:#007900}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i = 1; i < max_instances_ + 1; i++) {

    if(i == 1)
      myfile << "    <th>(Seg,Inst)</th>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <th>"
             << "(" << ids.first << "," << ids.second << ")" << "</th>"
             << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "  <tr>" << std::endl;

      if(instance_id_to_index_lookup_[i][j] == -1) {
        myfile << "  </tr>" << std::endl;
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <td>"
             << "(" << ids.first << "," << ids.second << ")" << "</td>"
             << std::endl;

      std::stringstream title;
      title << "title=\"(" << ids.first << "," << ids.second << ")";

      for(int k = 1; k < max_instances_ + 1; k++) {

        for(int l=0; l < number_of_segments_; l++) {

          if(instance_id_to_index_lookup_[k][l] == -1) {
            continue;
          }

          int row = instance_id_to_index_lookup_[i][j];
          int column = instance_id_to_index_lookup_[k][l];

          std::pair<int,int> column_ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[k][l]);

          std::stringstream title2;
          title2 << ", ("
                 << column_ids.first << "," << column_ids.second << ")\"";

          if(conflict_table_[row] [column]) {

            if(column_ids.second != ids.second) {

              if((column_ids.second + ids.second) % 2 == 1) {

                myfile << "    <td class=\"tg-red0\" " << title.str()
                       << title2.str()<< ">" << std::endl;
              } else {

                myfile << "    <td class=\"tg-red1\" " << title.str()
                       << title2.str()<< ">" << std::endl;
              }

            } else {

              myfile << "    <td class=\"tg-red2\" " << title.str()
                     << title2.str()<< ">" << std::endl;
            }

#if 1
            if(debugging_) {
              for(std::set<Conflict>::iterator
                  iter  = conflict_info_[row][column].
                              conflict_variables_.begin();
                  iter != conflict_info_[row][column].
                              conflict_variables_.end();
                  iter++) {

                // Here we want to print the conflict into the conflict table.
                // If the node_ value is equl to NULL, we have to use the
                // mapped variable. Otherwise, we use the expression node_
                if((*iter).node_ == NULL) {

                  myfile << "      <pre>"
                         << (*iter).variable_.symbol_->
                              get_declaration()->unparseToString()
                         << ", " << (*iter).variable_.instance_id_;

                  myfile << "</pre>"  << std::endl;

                } else {
                  myfile << "      <pre>"
                         << (*iter).node_->unparseToString()
                         << "</pre>"  << std::endl;
                }
              }
            }
#endif

            myfile << "    </td>" << std::endl;
          } else {

            if(column_ids.second != ids.second) {

              if((column_ids.second + ids.second) % 2 == 1) {

                myfile << "    <td class=\"tg-green0\" " << title.str()
                       << title2.str()<< "></td>" << std::endl;
              } else {

                myfile << "    <td class=\"tg-green1\" " << title.str()
                       << title2.str()<< "></td>" << std::endl;
              }
            } else {

              myfile << "    <td class=\"tg-green2\" " << title.str()
                     << title2.str()<< "></td>" << std::endl;
            }
          }
        }
      }

      myfile << "  </tr>" << std::endl;
    }
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

void
risc::sg::ConflictTable::print_lookup_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i=0; i < number_of_segments_ + 1; i++) {

    if(i == 0) {
      myfile << "    <th> </th>" << std::endl;

    } else {
      myfile << "    <th>"
             << instance_id_to_index_lookup_[0][i-1]
             << "</th>" << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i=1; i < max_instances_ + 1; i++) {

    myfile << "  <tr>" << std::endl;
    myfile << "<td>" << i-1 << "</td>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "<td>" << instance_id_to_index_lookup_[i][j]
             << "</td>" << std::endl;
    }
    myfile << "  </tr>" << std::endl;
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

risc::sg::VertexDescriptor
risc::sg::ConflictTable::segment_id_to_vertex_descriptor(int id)
{
  for(int i = 0; i < number_of_segments_; i++) {
    if(instance_id_to_index_lookup_[0][i] == id) {
      return vertex_lookup_[i];
    }
  }

  assert("segment_id_to_vertex_descriptor: Id doesn't exist" && 0);
}

int risc::sg::ConflictTable::get_max_instances(int segment_id)
{
  for(int i=0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id)  {

      int counter = 1;
      for(; counter < max_instances_ + 1; counter++) {
        if(instance_id_to_index_lookup_[counter][i] == -1) {
          break;
        }
      }

      return counter -1;
    }
  }

  assert("Cannot determine max instances" && 0);
  return -1;
}

int
risc::sg::ConflictTable::get_channel_id(
    const SymbolWithPath &swp,
    int instance,
    bool dynamic_analysis)
{
  if(dynamic_analysis) {
    return get_channel_id_dynamically(swp, instance);

  } else {
    return get_channel_id_statically(swp, instance);
  }
}

int
risc::sg::ConflictTable::get_channel_id_statically(
    const SymbolWithPath &swp,
    int instance)
{
  // this variable is declared in a channel and the channel has its own thread
  if(swp.pcp_.path_.size() == 0 && swp.pcp_.port_ == NULL) {
    return instance;
  }
  
  if(swp.pcp_.pcp_type_==0){
    SgVariableDefinition *port_var
        = tools::get_port_variable(swp.pcp_.path_.front());
    return risc::get_design()->port_mapping_cache_
              .mapping_[instance][port_var].second;
  }else{
  //change 2018/7/18
    SgVariableDefinition *port_var
        = swp.pcp_.port_;
    
    return risc::get_design()->port_mapping_cache_
              .mapping_[instance][port_var].second;

  }
}

int
risc::sg::ConflictTable::get_channel_id_dynamically(
    const SymbolWithPath &swp,
    int instance)
{
  // this variable is declared in a channel and the channel has its own thread
  if(swp.pcp_.path_.size() == 0) {
    return instance;
  }

  SgVariableDefinition *port_var
    = tools::get_port_variable(swp.pcp_.path_.front());

  SgVariableDeclaration *port_decl = port_var->get_declaration();

  std::string port_name(SageInterface::get_name(port_decl));

  // deleting __variable_declaration__ out of the string
  port_name.erase(0, 22);

  int counter = 0;
  DynamicModule *hosting_module = get_module_instance(
      path_instance_mapper_->get_dynamic_tree_root(),
      SageInterface::get_name(
        SageInterface::getEnclosingClassDefinition(port_decl)),
      instance,
      counter);

  DynamicPort *dyn_port = NULL;
  for(std::list<DynamicPort*>::iterator
      port_iter  = hosting_module->ports_.begin();
      port_iter != hosting_module->ports_.end();
      port_iter++) {
    if((*port_iter)->name_ == port_name) {
      dyn_port = *port_iter;
    }
  }

  int id = get_mapped_channel_id_of_port(dyn_port);

  return id;
}

void risc::sg::ConflictTable::print_conflicts_per_instance()
{
  if(!debugging_ || path_instance_mapper_->perform_dynamic_analysis()) {
    std::cout << "The function ConflictTable::print_conflicts_per_instance() "
              << "is only available in the debugging mode "
              << "in combination with the static analysis."
              << std::endl;
    return;
  }
  std::cout.width(5);
  std::cout << std::right << "";
  for(int i = 0; i < max_instances_; i++) {
    std::cout.width(5);
    std::cout << std::right << i;
  }
  std::cout << std::endl;

  for(int i = 0; i < max_instances_; i++) {


    std::cout.width(5);
    std::cout << std::right << i;

    for(int j = 0; j < max_instances_; j++) {

      std::cout.width(5);
      std::cout << std::right
                << conflicts_per_instance_[i * max_instances_ + j];
    }
    std::cout << std::endl;
  }
  std::cout << std::endl;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
