#ifndef COMMAND_LINE_PARSING_H_INCLUDED_
#define COMMAND_LINE_PARSING_H_INCLUDED_

#include <vector>
#include <string>

namespace risc {

class Design;

namespace tools {


/**
 * \brief      This function takes the parameter of the command line and
 *             creates a source file which includes all the source files.
 *             Afterwards all source files are replaced with new file in the
 *             command parameters. Example
 *
 *             Input:
 *               ./myprogram file_a.cpp file_b.cpp file_c.cpp
 *
 *             Output
 *               ./myprogram __main__.cpp
 *
 *             New file __main__.cpp
 *               #include "file_a.cpp"
 *               #include "file_b.cpp"
 *               #include "file_c.cpp"
 *
 * \return     vector of strings with the new parameters
 */
std::vector<std::string>
preparsing_commandline_arguments(int &argc, char **&argv,
    std::vector<std::string> &input_files);

/**
 * \brief  This function searches for the command line arguments for
 *         the dynamic analysis, namely, -inst:dir_output_filename <filename>.
 *         ONLY FOR INTERNAL USE!
 *         MODIFIES THE COMMAND LINE PARAMETERS
 *
 * \return The return string will be empty if the dyn flag is not set.
 *         Otherwise, the string contains the filename
 */
std::string instrumentor_output_filename(int &argc, char **&argv);

/**
 * \brief  This function searches for the command line arguments for
 *         the dynamic analysis, namely, -risc:elab <filename>.
 *         ONLY FOR INTERNAL USE!
 *         MODIFIES THE COMMAND LINE PARAMETERS
 *
 * \return The return string will be empty if the dyn flag is not set.
 *         Otherwise, the string contains the filename
 */
std::string dynamic_analysis_input_filename(int &argc, char **&argv);

/**
 *  \class ParseStatus
 *  \brief This class stores the command-line options
 */
class ParseStatus
{
  public:
    /**
     * \brief Verbosity level in RISC library and tools
     */
    unsigned int verbosityLevel;

    /**
     * \brief Warning level in RISC library and tools
     */   
    unsigned int warningLevel;

    /**
     * \brief Backend compiler name
     */
    std::string compilerName;
    
    std::vector < std::string > input_psg_file_names;
    std::string output_psg_file_name;
    
    /**
     * \brief Flag to compile for Intel MIC architecture
     */ 
    bool mic;

    ParseStatus();
    ParseStatus& operator=(const ParseStatus& copy);
};

/**
 * \fn ParseCommandlineOptions
 * \brief Parse command-line options that are related to RISC library
 */
void
ParseCommandlineOptions(int &argc, char** &argv, ParseStatus* status);

/**
 * \fn AddCommandlineOption
 * \brief Add a new command-line option to argc & argv
 */
void
AddCommandlineOption(int &argc, char** &argv, std::string option);

/**
 * \fn AddCommandlineOption
 * \brief Add a new command-line option to the command-line argument
 *        list in SgProject 
 */
void
AddCommandlineOption(risc::Design &design, std::string option);

/**
 * \fn RemoveCommandlineOption
 * \brief Remove a command-line option from the command-line argument
 *        list in SgProject
 */
void
RemoveCommandlineOption(risc::Design &design, std::string option);

} // end namespace tools

} // end namespace risc

#endif /* COMMAND_LINE_PARSING_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
