// particle_channel4r.cc: regression test for event synchronization
// (originated from NoC particle simulation example with lost events)
// same as particle_channel2r.cc, but with duplicated channel code
// RD, 04/19/17

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS (42*10)
#define WORKLOAD   1000000

// deterministic MT-safe pseudo random number in range [0...Max]
// (based on example by 'man rand', 4/8/17, RD)
int Rnd(unsigned long &Next, int Max)
{
    Next = Next * 1103515245 + 12345;
    return((unsigned)(Next/65536) % (Max+1));
}

// simulated workload

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

// user-defined data type

typedef struct Particle_t
{
  double 	x, y;

  inline friend std::ostream& operator<<(
	std::ostream& os, const Particle_t& co)
  { os << co.x << ' ' << co.y << '\n';
    return os;
  }
} particle_t;

// user-defined typed double-handshake channel (with transceiver interface)

class i_particle_transceiver: virtual public sc_interface
{
    public:	// (mimic a subset of sc_fifo interface, without references)
	virtual particle_t read(void) = 0;
	virtual void write(const particle_t) = 0;
};

class c_particle_double_handshake_channel:
	public sc_channel, public i_particle_transceiver
{
    public:
	c_particle_double_handshake_channel(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_particle_double_handshake_channel()
	{ }

	particle_t read(void)
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    particle_t temp = tmpd;
	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	void write(const particle_t X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}

    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	particle_t tmpd;
};
class c_particle_double_handshake_channel2 /* DUPLICATE OF ABOVE */:
	public sc_channel, public i_particle_transceiver
{
    public:
	c_particle_double_handshake_channel2(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_particle_double_handshake_channel2()
	{ }

	particle_t read(void)
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    particle_t temp = tmpd;
	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	void write(const particle_t X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}

    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	particle_t tmpd;
};

// Ping

SC_MODULE(Ping)
{
    sc_port<i_particle_transceiver> Port;
    unsigned Count;
    unsigned long w;

    void main(void);

    SC_CTOR(Ping)
    {
	Count = 0;
	SC_HAS_PROCESS(Ping);
	SC_THREAD(main);
    }
};

void Ping::main(void)
{
    particle_t p;
    int i;
    unsigned long RndNext = 1;

    for(i=0; i<ITERATIONS; i++)
    {
	p.x = i;
	p.y = i;
	Port->write(p);
	Count++;
	w = Work(Rnd(RndNext, WORKLOAD));
//	wait(1, SC_NS);
	p = Port->read();
	assert(p.x == -i);
	assert(p.y == -i);
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Pong

SC_MODULE(Pong)
{
    sc_port<i_particle_transceiver> Port;
    unsigned Count;
    unsigned long w;

    void main(void);

    SC_CTOR(Pong)
    {
	Count = 0;
	SC_HAS_PROCESS(Pong);
	SC_THREAD(main);
    }
};

void Pong::main(void)
{
    particle_t p;
    int i;
    unsigned long RndNext = 2;

    for(i=0; i<ITERATIONS; i++)
    {
	p = Port->read();
	assert(p.x == i);
	assert(p.y == i);
	p.x *= -1.0;
	p.y *= -1.0;
	w = Work(Rnd(RndNext, WORKLOAD));
//	wait(1, SC_NS);
	Port->write(p);
	Count++;
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Top

SC_MODULE(TOP)
{
    c_particle_double_handshake_channel c1;
    c_particle_double_handshake_channel2 c2;
    Ping ping1, ping2;
    Pong pong1, pong2;

    SC_CTOR(TOP)
    :c1("c1"), ping1("ping1"), pong1("pong1"),
     c2("c2"), ping2("ping2"), pong2("pong2")
    {
    }

    void before_end_of_elaboration()
    {
        ping1.Port.bind(c1);
        pong1.Port.bind(c1);

        ping2.Port.bind(c2);
        pong2.Port.bind(c2);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("Ping-pong over 2 duplicated particle channels starts\n");
    sc_start();
    assert(top.ping1.Count == ITERATIONS);
    assert(top.pong1.Count == ITERATIONS);
    assert(top.ping2.Count == ITERATIONS);
    assert(top.pong2.Count == ITERATIONS);
    printf("Ping-pong over 2 duplicated particle channels done.\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
